from __future__ import annotations

import argparse
import base64
import contextlib
from collections.abc import Iterable
from dataclasses import dataclass, field
import functools
import json
import logging
import os
import pkg_resources
import sys
import textwrap
import time
from typing import Any, Callable, Dict, Generator, cast, Optional, List, Union, Tuple


from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor

from io import StringIO
from pubtools.pluggy import pm

from pubtools.tracing import get_trace_wrapper

tw = get_trace_wrapper()
LOG = logging.getLogger("pubtools.quay")

INTERNAL_DELIMITER = "----"
MAX_RETRY_WAIT = 120


@dataclass
class FData:
    """Dataclass for holding data for a function execution.

    Args:
        args (Iterable[Any]): Arguments for the function.
        kwargs (Dict[str, Any]): Keyword arguments for the function.
    """

    args: Iterable[Any]
    kwargs: Dict[str, Any] = field(default_factory=dict)


def run_in_parallel(func: Callable[..., Any], data: List[Any], threads: int = 10) -> Dict[Any, Any]:
    """Run method on data in parallel.

    Args:
        func (function): Function to run on data
        data (list): List of tuples which are used as arguments for the function
    Returns:
        dict: List of result in the same order as data.
    """
    future_results = {}
    results = {}
    with ThreadPoolExecutor(max_workers=threads) as executor:
        future_results = {
            executor.submit(func, *data_entry.args, **data_entry.kwargs): n
            for n, data_entry in enumerate(data)
        }
        for future in futures.as_completed(future_results):
            if future.exception() is not None:
                raise cast(BaseException, future.exception())
            results[future_results[future]] = future.result()
    return dict(sorted(results.items(), key=lambda kv: kv[0]))


def setup_arg_parser(args: Dict[Any, Any]) -> argparse.ArgumentParser:
    """
    Set up ArgumentParser with the provided arguments.

    Args:
        args (dict)
            Dictionary of argument aliases and options to be consumed by ArgumentParser.
    Returns:
        (ArgumentParser) Configured instance of ArgumentParser.
    """
    parser = argparse.ArgumentParser()
    arg_groups: Dict[Any, Any] = {}
    for aliases, arg_data in args.items():
        holder = parser
        if "group" in arg_data:
            arg_groups.setdefault(arg_data["group"], parser.add_argument_group(arg_data["group"]))
            holder = arg_groups[arg_data["group"]]
        action = arg_data.get("action")
        if not action and arg_data["type"] == bool:
            action = "store_true"
        kwargs = {
            "help": arg_data.get("help"),
            "required": arg_data.get("required", False),
            "default": arg_data.get("default"),
        }
        if action:
            kwargs["action"] = action
        else:
            kwargs["type"] = arg_data.get("type", "str")
            kwargs["nargs"] = arg_data.get("count")

        holder.add_argument(*aliases, **kwargs)

    return parser


def add_args_env_variables(
    parsed_args: argparse.Namespace, args: Dict[Any, Any]
) -> argparse.Namespace:
    """
    Add argument values from environment variables.

    Args:
        parsed_args ():
            Parsed arguments object.
        args (dict):
            Argument definition.
    Returns:
        Modified parsed arguments object.
    """
    for aliases, arg_data in args.items():
        named_alias = [x.lstrip("-").replace("-", "_") for x in aliases if x.startswith("--")][0]
        if arg_data.get("env_variable"):
            if not getattr(parsed_args, named_alias) and os.environ.get(arg_data["env_variable"]):
                setattr(parsed_args, named_alias, os.environ.get(arg_data["env_variable"]))
    return parsed_args


@contextlib.contextmanager
def capture_stdout() -> Generator[StringIO, None, None]:
    """Capture sys.stdout to stream buffer."""
    old_stdout = sys.stdout
    sys.stdout = new_stdout = StringIO()

    try:
        yield new_stdout
    finally:
        sys.stdout = old_stdout


@contextlib.contextmanager
def setup_entry_point_cli(
    entry_tuple: Tuple[str, str, str],
    name: Optional[str],
    args: List[str],
    environ_vars: Dict[str, Any],
) -> Generator[Callable[[], Any], None, None]:
    """
    Set up an entrypoint as a context manager.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.
    """
    orig_environ = os.environ.copy()

    try:
        # First argv element is always the entry point name.
        # For a console_scripts entry point, this will be the same value
        # as if the script was invoked directly. For any other kind of entry point,
        # this value is probably meaningless.
        for key in environ_vars:
            os.environ[key] = environ_vars[key]
        entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
        if args:
            if name:
                func_args = [name]
            else:
                func_args = []
            func_args.extend(args)
            if entry_tuple[1] == "console_scripts":
                yield functools.partial(entry_point_func, func_args)
            else:
                yield functools.partial(entry_point_func, *func_args)
        else:
            yield entry_point_func
    finally:
        os.environ.update(orig_environ)

        to_delete = [key for key in os.environ if key not in orig_environ]
        for key in to_delete:
            del os.environ[key]


def run_entrypoint(
    entry_tuple: Tuple[str, str, str],
    name: Optional[str],
    args: List[str],
    environ_vars: Dict[str, Any],
    capture_out: bool = True,
) -> Any:
    """
    Run an entrypoint function and return its return value.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.
        capture_out (bool):
            Whether to capture stdout.

    Returns (str):
        Data returned by the entrypoint.
    """
    raw_args = " ".join([entry_tuple[2]] + args)
    wrapped_lines = textwrap.wrap(
        raw_args, 100, subsequent_indent="  ", break_on_hyphens=False, break_long_words=False
    )

    LOG.info("Running task with arguments:")
    for idx, line in enumerate(wrapped_lines):
        suffix = ""
        if idx != len(wrapped_lines) - 1:
            # shell-style backslash to indicate continuation
            suffix = " \\"
        LOG.info("%s%s", line, suffix)

    with setup_entry_point_cli(entry_tuple, name, args, environ_vars) as entry_func:
        if capture_out:
            with capture_stdout():
                pyret = entry_func()
        else:
            pyret = entry_func()

    return pyret


def get_internal_container_repo_name(external_name: str) -> str:
    """
    Transform a repository name to an internal form in which it exists on Quay.io.

    Expected input format: <namespace>/<product>
    Generated output format: <namespace>----<product>

    NOTE: Repositories without a delimeter "/" may actually exist. In that case, the function
    simply returns the repo without any alterations.

    Args:
        external_name (str):
            External repository name.
    Returns:
        Internal repository name.
    """
    if external_name.count("/") == 0:
        return external_name

    if external_name.count("/") > 1 or external_name[0] == "/" or external_name[-1] == "/":
        raise ValueError(
            "Input repository containing a delimeter should "
            "have the format '<namespace>/<product>'",
            external_name,
        )

    return external_name.replace("/", INTERNAL_DELIMITER)


def get_external_container_repo_name(internal_name: str) -> str:
    """
    Transform a repository name to an external form in which it's visible to customers.

    Expected input format: <namespace>----<product>
    Generated output format: <namespace>/<product>

    NOTE: Repositories without a delimeter "----" may actually exist. In that case, the function
    simply returns the repo without any alterations.

    Args:
        internal_name (str):
            Internal repository name.
    Returns:
        External repository name.
    """
    if internal_name.count(INTERNAL_DELIMITER) == 0:
        return internal_name

    if (
        internal_name.count(INTERNAL_DELIMITER) > 1
        or internal_name.find(INTERNAL_DELIMITER) == 0
        or internal_name.find(INTERNAL_DELIMITER) == len(internal_name) - 4
    ):
        raise ValueError(
            "Input repository containing a delimeter should "
            "have the format '<namespace>----<product>'",
            internal_name,
        )

    return internal_name.replace(INTERNAL_DELIMITER, "/")


def task_status(event: str) -> Dict[str, Dict[str, str]]:
    """Helper function. Expand as necessary."""  # noqa: D401
    return dict(event={"type": event})


def log_step(step_name: str) -> Callable[[Any], Any]:
    """
    Log status for methods which constitute an entire task step.

    Args:
        step_name (str):
            Name of the task step, e.g., "Tag images".
    """
    event_name = step_name.lower().replace(" ", "-")

    def decorate(fn: Callable[..., Any]) -> Callable[..., Any]:
        # Add instrumentation trace for all push steps.
        fn = tw.instrument_func(span_name=event_name)(fn)

        @functools.wraps(fn)
        def fn_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                LOG.info("%s: Started", step_name, extra=task_status("%s-start" % event_name))
                ret = fn(*args, **kwargs)
                LOG.info("%s: Finished", step_name, extra=task_status("%s-end" % event_name))
                return ret
            except Exception:
                LOG.error("%s: Failed", step_name, extra=task_status("%s-error" % event_name))
                raise

        return fn_wrapper

    return decorate


def get_pyxis_ssl_paths(target_settings: Dict[str, Any]) -> Tuple[str, str]:
    """
    Get certificate and key paths for Pyxis SSL authentication.

    First attempt is made by invoking the hook implementation 'get_cert_key_paths'.
    If nothing is returned (no hook implementation is registered), fallback on target settings
    values of 'pyxis_ssl_cert' and 'pyxis_ssl_key'. If multiple values are returned (multiple
    hook implementations are registered), take the first response (from the implementation which
    was registered last). Otherwise, raise an error.

    Args:
        target_settings (dict):
            Dictionary containing various task-related settings.
    Returns ((str, str)):
        Paths to Pyxis SSL certificate and key.
    """
    result = pm.hook.get_cert_key_paths(server_url=target_settings["pyxis_server"])
    if result:
        cert, key = result
    elif "pyxis_ssl_cert" not in target_settings or "pyxis_ssl_key" not in target_settings:
        raise ValueError(
            "No key and certificate paths were provided for Pyxis SSL authentication. "
            "Please, either provide hook implementation for 'get_cert_key_paths_plugin' or "
            "specify 'pyxis_ssl_cert' and 'pyxis_ssl_key' in the target settings."
        )
    else:
        cert = target_settings["pyxis_ssl_cert"]
        key = target_settings["pyxis_ssl_key"]

    return (cert, key)


def timestamp() -> str:
    """Return now() timestamp."""
    return str(time.time()).split(".")[0]


def run_with_retries(
    function: Callable[[], Any], message: str, tries: int = 4, wait_time_increase: int = 10
) -> Any:
    """
    Run the specified function until it succeeds or maximum retries are reached.

    Wait time will increase after every retry, up to a point defined by MAX_RETRY_WAIT.

    Args:
        function (callable):
            Function that should be retried. It must be able to run with 0 parameters.
        message (str):
            Message describing the action performed by the function. For example, "tag images".
        tries (int):
            Numbers of times to run the function before giving up.
        wait_time_increase (int):
            Time increase (in seconds) to wait before running the function again. Example (default):
            RUN -> WAIT 0 -> RUN -> WAIT 10 -> RUN -> WAIT 20 -> RUN
    """
    wait_time = 0
    for i in range(tries):
        try:
            result = function()
            if i != 0:
                LOG.info("%s succeeded [try: %s/%s]" % (message, i + 1, tries))
            return result
        except Exception as e:
            if i < tries - 1:
                wait_time = i * wait_time_increase if wait_time < MAX_RETRY_WAIT else MAX_RETRY_WAIT
                LOG.warning(
                    "%s failed. Will retry in %d seconds [try %s/%s]: %s"
                    % (message, wait_time, i + 1, tries, str(e))
                )
                time.sleep(wait_time)
                continue
            LOG.error("%s repeatedly fails" % message)
            raise


def retry(
    message: str, tries: int = 4, wait_time_increase: int = 10
) -> Callable[[Callable[..., Any]], Any]:
    """
    Retry decorated function.

    Args:
        message (str):
            Message describing the action performed by the function. For example, "tag images".
        tries (int):
            Numbers of times to run the function before giving up.
        wait_time_increase (int):
            Time increase (in seconds) to wait before running the function again. Example (default):
            RUN -> WAIT 0 -> RUN -> WAIT 10 -> RUN -> WAIT 20 -> RUN
    """

    def inner_retry(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper_func(*args: Any, **kwargs: Any) -> Any:
            bound = functools.partial(func, *args, **kwargs)
            return run_with_retries(bound, message, tries, wait_time_increase)

        return wrapper_func

    return inner_retry


def parse_index_image(build_details: Any) -> Tuple[str, str, str]:
    """
    Get registry, namespace and repository of a resolved internal index image.

    Args:
        build_details (dict):
            Dictionary of IIB build details.
    Returns ((str, str, str)):
        Registry, namespace, repository of an image.
    """
    image_path = build_details.internal_index_image_copy_resolved.split("@")[0]
    registry, namespace, repo = image_path.split("/")

    return (registry, namespace, repo)


def get_basic_auth(host: str) -> Union[Tuple[Optional[str], Optional[str]], List[str]]:
    """
    Look for container config file for username and password.

    Args:
        host (str):
            Hostname of a container registry.
    Returns ((str, str)):
        Username, password of a registry.
    """
    home_dir = os.path.expanduser("~")
    conf_file = os.path.join(home_dir, ".docker/config.json")
    if os.path.isfile(conf_file):
        with open(conf_file) as f:
            config = json.load(f)
        auth = config.get("auths", {}).get(host, {}).get("auth")
        if auth:
            return base64.b64decode(auth).decode().split(":")
    return None, None


def pyxis_get_repo_metadata(repo: str, target_settings: Dict[str, Any]) -> Any:
    """
    Invoke the 'get-repo-metadata' entrypoint from pubtools-pyxis.

    Args:
        repo (str):
            Repository to get the metadata of.
        target_settings (dict):
            Settings used for setting the values of the entrypoint parameters.

    Returns (dict):
        Parsed response from Pyxis.
    """
    cert, key = get_pyxis_ssl_paths(target_settings)

    args = ["--pyxis-server", target_settings["pyxis_server"]]
    args += ["--pyxis-ssl-crtfile", cert]
    args += ["--pyxis-ssl-keyfile", key]
    args += ["--repo-name", repo]

    env_vars: Dict[Any, Any] = {}
    metadata = run_entrypoint(
        ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-repo-metadata"),
        "pubtools-pyxis-get-repo-metadata",
        args,
        env_vars,
    )
    return metadata


def set_aws_kms_environment_variables(target_settings: Dict[str, Any], profile_name: str) -> None:
    """
    Set environment variables required to use an AWS KMS key.

    The values are set from target settings based on the selected profile. Multiple profiles are
    supported as multiple keys may be used during a push. Following env variables are set:
    AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_DEFAULT_REGION. Missing data in target settings
    will not result in an error. Target settings is expected to be in the following format:
    {
    "aws_kms_credentials": {
        "profile-1": {
            "aws_access_key_id": "id1",
            "aws_secret_access_key": "key1",
            "aws_default_region": "us-east-1"
            },
        "profile-2": {
            "aws_access_key_id": "id2",
            "aws_secret_access_key": "key2",
            "aws_default_region": "us-east-2"
            }
        },
    "other_target_setting":"data"
    }

    Args:
        target_settings (dict):
            Target settings.
        profile_name (str):
            Profile name whose credentials to apply.
    """
    if "aws_kms_credentials" not in target_settings:
        LOG.warning(
            "Target settings are missing the aws_kms_credentials property, "
            "cannot set AWS KMS environment variables"
        )
        return
    if profile_name not in target_settings["aws_kms_credentials"]:
        LOG.warning(f"AWS KMS profile {profile_name} not found in the target settings")
        return

    if "aws_access_key_id" not in target_settings["aws_kms_credentials"][profile_name]:
        LOG.warning(
            "Cannot set AWS KMS environment variable AWS_ACCESS_KEY_ID, "
            f"value missing in profile {profile_name}"
        )
    else:
        os.environ["AWS_ACCESS_KEY_ID"] = target_settings["aws_kms_credentials"][profile_name][
            "aws_access_key_id"
        ]

    if "aws_secret_access_key" not in target_settings["aws_kms_credentials"][profile_name]:
        LOG.warning(
            "Cannot set AWS KMS environment variable AWS_SECRET_ACCESS_KEY, "
            f"value missing in profile {profile_name}"
        )
    else:
        os.environ["AWS_SECRET_ACCESS_KEY"] = target_settings["aws_kms_credentials"][profile_name][
            "aws_secret_access_key"
        ]

    if "aws_default_region" not in target_settings["aws_kms_credentials"][profile_name]:
        LOG.warning(
            "Cannot set AWS KMS environment variable AWS_DEFAULT_REGION, "
            f"value missing in profile {profile_name}"
        )
    else:
        os.environ["AWS_DEFAULT_REGION"] = target_settings["aws_kms_credentials"][profile_name][
            "aws_default_region"
        ]



import mock
import pytest

from pubtools._quay import clear_repo
from tests.utils.misc import GetManifestSideEffect


@mock.patch("pubtools._quay.clear_repo.clear_repositories")
def test_arg_constructor_required_args(mock_clear_repositories):
    required_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]
    clear_repo.clear_repositories_main(required_args)
    repositories, settings = mock_clear_repositories.call_args[0]
    assert repositories == "namespace/image"
    assert settings["quay_org"] == "quay-organization"
    assert settings["quay_user"] == "some-user"
    assert settings["quay_password"] == "some-password"
    assert settings["quay_api_token"] == "some-token"
    assert settings["pyxis_server"] == "pyxis-url.com"
    assert settings["pyxis_ssl_crtfile"] == "/path/to/file.crt"
    assert settings["pyxis_ssl_keyfile"] == "/path/to/file.key"
    assert settings["quay_host"] == "quay-host.com"


@mock.patch.dict("os.environ", {"QUAY_API_TOKEN": "api_token", "QUAY_PASSWORD": "some-password"})
@mock.patch("pubtools._quay.clear_repo.clear_repositories")
def test_arg_constructor_all_args(mock_clear_repositories):
    all_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]
    clear_repo.clear_repositories_main(all_args)
    repositories, settings = mock_clear_repositories.call_args[0]

    assert repositories == "namespace/image"
    assert settings["quay_org"] == "quay-organization"
    assert settings["quay_user"] == "some-user"
    assert settings["quay_password"] == "some-password"
    assert settings["pyxis_server"] == "pyxis-url.com"
    assert settings["pyxis_ssl_crtfile"] == "/path/to/file.crt"
    assert settings["pyxis_ssl_keyfile"] == "/path/to/file.key"
    assert settings["quay_api_token"] == "api_token"
    assert settings["quay_host"] == "quay-host.com"


@mock.patch("pubtools._quay.clear_repo.clear_repositories")
def test_args_missing_repositories(mock_clear_repositories):
    wrong_args = [
        "dummy",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
    ]

    with pytest.raises(SystemExit) as system_error:
        clear_repo.clear_repositories_main(wrong_args)

    assert system_error.type == SystemExit
    assert system_error.value.code == 2
    mock_clear_repositories.assert_not_called()


@mock.patch("pubtools._quay.clear_repo.clear_repositories")
def test_args_missing_api_token(mock_clear_repositories):
    wrong_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]

    with pytest.raises(ValueError, match="--quay-api-token must be specified"):
        clear_repo.clear_repositories_main(wrong_args)

    mock_clear_repositories.assert_not_called()


@mock.patch("pubtools._quay.clear_repo.clear_repositories")
def test_args_missing_quay_password(mock_clear_repositories):
    wrong_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]

    with pytest.raises(ValueError, match="--quay-password must be specified"):
        clear_repo.clear_repositories_main(wrong_args)

    mock_clear_repositories.assert_not_called()


@mock.patch("pubtools._quay.clear_repo.untag_images")
@mock.patch("pubtools._quay.clear_repo.QuayClient")
def test_run(
    mock_quay_client,
    mock_untag_images,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    signer_wrapper_remove_signatures,
):
    args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--signers",
        "msg_signer",
        "--quay-host",
        "quay-host.com",
    ]
    mock_get_repo_tags = mock.MagicMock()
    mock_get_repo_tags.return_value = {"tags": ["1", "2"]}
    mock_quay_client.return_value.get_repository_tags = mock_get_repo_tags
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:3333333333",
            "reference": "registry.io/namespace/image:1",
            "sig_key_id": "key",
            "repository": "namespace/image",
        },
    ]
    mock_quay_client.return_value.get_manifest.side_effect = GetManifestSideEffect(
        v2s1_manifest, src_manifest_list, call_limit=16
    )
    clear_repo.clear_repositories_main(args)

    mock_quay_client.assert_called_once_with("some-user", "some-password")
    mock_get_repo_tags.assert_called_with("quay-organization/namespace----image")
    signer_wrapper_remove_signatures.assert_called_with(
        [1],
    )
    mock_untag_images.assert_called_once_with(
        [
            "quay.io/quay-organization/namespace----image:1",
            "quay.io/quay-organization/namespace----image:2",
        ],
        "some-token",
        remove_last=True,
        quay_user="some-user",
        quay_password="some-password",
    )


@mock.patch("pubtools._quay.clear_repo.untag_images")
@mock.patch("pubtools._quay.clear_repo.QuayClient")
def test_run_multiple_repos(
    mock_quay_client,
    mock_untag_images,
    hookspy,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    signer_wrapper_remove_signatures,
):
    args = [
        "dummy",
        "--repositories",
        "namespace/image,namespace/image2",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--signers",
        "msg_signer",
        "--quay-host",
        "quay-host.com",
    ]
    mock_get_repo_tags = mock.MagicMock()
    mock_get_repo_tags.side_effect = [
        {"tags": ["1", "2"]},
        {"tags": ["3", "4"]},
        {"tags": ["1", "2"]},
        {"tags": ["3", "4"]},
    ]
    mock_quay_client.return_value.get_repository_tags = mock_get_repo_tags
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:5555555555",
            "reference": "registry.io/namespace/image2:3",
            "sig_key_id": "key",
            "repository": "namespace/image2",
        },
        {
            "_id": 2,
            "manifest_digest": "sha256:1111111111",
            "reference": "registry.io/namespace/image2:3",
            "sig_key_id": "key",
            "repository": "namespace/image2",
        },
    ]

    mock_quay_client.return_value.get_manifest.side_effect = GetManifestSideEffect(
        v2s1_manifest, src_manifest_list, call_limit=20
    )

    clear_repo.clear_repositories_main(args)

    mock_quay_client.assert_called_once_with("some-user", "some-password")
    assert mock_get_repo_tags.call_count == 4
    assert mock_get_repo_tags.call_args_list[0] == mock.call("quay-organization/namespace----image")
    assert mock_get_repo_tags.call_args_list[1] == mock.call(
        "quay-organization/namespace----image2"
    )
    assert mock_get_repo_tags.call_args_list[2] == mock.call("quay-organization/namespace----image")
    assert mock_get_repo_tags.call_args_list[3] == mock.call(
        "quay-organization/namespace----image2"
    )
    signer_wrapper_remove_signatures.assert_called_with(
        [1, 2],
    )

    mock_untag_images.assert_called_once_with(
        [
            "quay.io/quay-organization/namespace----image2:3",
            "quay.io/quay-organization/namespace----image2:4",
            "quay.io/quay-organization/namespace----image:1",
            "quay.io/quay-organization/namespace----image:2",
        ],
        "some-token",
        remove_last=True,
        quay_user="some-user",
        quay_password="some-password",
    )

    # Should have activated these hooks.
    assert hookspy == [
        ("task_start", {}),
        ("get_cert_key_paths", {"server_url": "pyxis-url.com"}),
        ("quay_repositories_cleared", {"repository_ids": ["namespace/image", "namespace/image2"]}),
        ("task_stop", {"failed": False}),
    ]



# this code has been copied from rcm-pub (pubd/lib/docker_signature.py)
import logging
import os
import json
from typing import Any, Callable, Sequence, Mapping, cast, Optional, List, Dict
import monotonic
import proton
from proton.handlers import MessagingHandler
from proton.reactor import Container

LOG = logging.getLogger("pubtools.quay")

# There are some linting errors since this was copied from rcm-pub
# flake8: noqa: D101, D107, D102, D205, D400, D403


class MessageHandlerTimeoutException(Exception):
    pass


class AMQPError(Exception):
    """Base class for AMQP-specific errors."""

    pass


class UMBSettings(object):
    __slots__ = [
        "broker_urls",
        "radas_address",
        "pub_cert",
        "ca_cert",
        "signing_timeout",
        "signing_throttle",
        "signing_retry",
    ]

    def __init__(
        self,
        broker_urls: List[str],
        radas_address: str = "VirtualTopic.eng.robosignatory.container.sign",
        pub_cert: str = "/etc/pub/umb-pub-cert-key.pem",
        ca_cert: str = "/etc/pki/tls/certs/ca-bundle.crt",
        signing_timeout: int = 600,
        signing_throttle: int = 100,
        signing_retry: int = 3,
    ) -> None:
        self.broker_urls = broker_urls
        self.radas_address = radas_address
        self.pub_cert = pub_cert
        self.ca_cert = ca_cert
        self.signing_timeout = signing_timeout
        self.signing_throttle = signing_throttle
        self.signing_retry = signing_retry


class AMQPEndpointError(AMQPError):
    """Signal an error with one of the AMQP endpoints.

    Parameters:
        endpoint_name (:py:obj:`str`): The endpoint at which the error
            occurred. Expected value is one of 'Connection', 'Session',
            'Link' or 'Transport'. The value provided will be title-cased,
            by the constructor so any "ALL UPPER" or "all lower" strings
            will be converted to "All Upper" or "All Lower".
        error_label (:py:obj:`str`): The name or label of the error that
            occurred e.g. `amqp:resource-limit-exceeded`. This class does
            not verify that the provided name matches expected error types
            from the AMQP specification. If no name is provided a default of
            'Unknown' is used.
        description (:py:obj:`str`): The description of the error that occurred.
            If no description is provided a default of 'No error description provided.'
            is used.
    """

    def __init__(
        self,
        endpoint_name: str,
        error_label: str = "Unknown",
        description: str = "No error description provided.",
    ) -> None:
        msg = "({endpoint!s}) {name}: {description}".format(
            endpoint=endpoint_name, name=error_label, description=description
        )
        super(AMQPEndpointError, self).__init__(msg)

    def __str__(self) -> str:
        # The message without the context of AMQPEndpointError doesn't mean much.
        return repr(self)


def _get_endpoint_error_condition(endpoint: proton.Endpoint) -> Optional[proton.Condition]:
    """Return the error condition on this endpoint if there is one.

    If there's no local error condition, return the remote error
    condition if there is one.

    Parameters:
        :obj:`proton.Endpoint`: The Proton Endpoint from which the error
            condition is being retrieved.

    Returns:
        :obj:`Optional[proton.Condition]`: The local endpoint error condition
            if there is one, or the remote endpoint error condition if there
            is one.
    """
    return endpoint.condition or endpoint.remote_condition


def do_nothing(*_: Any, **__: Any) -> None:
    """A convenient placeholder function that does, as advertised, nothing."""
    pass  # pragma: no cover


def raise_error(error: Any) -> None:
    """A convenient placeholder function that raises an error."""
    raise error  # pragma: no cover


class ManifestClaimsHandler(MessagingHandler):  # type: ignore
    """
    The manifest claims handler waits for messages from the UMB
    and will check against the manifest claims sent until all claims are received.

    Parameters:
        settings (:obj:`UMBSettings`): A UMBSettings object prescribing the
            correct settings for the remote UMB peer through which communication
            with Radas happens.
        claim_messages (:obj:`Sequence[Mapping[str, Any]]`): The manifest claim
            messages that need to be signed by Radas.
        message_sender_callback (:obj:`Callable[[Sequence[Mapping[str, Any]]], None]:
            A callback that will be invoked with to send messages to Radas for
            signing.
        on_message_callback (:obj:`Callable[[Mapping[str, Any]], None]:
            Callback invoked if an incoming response is matched with
            with a request. The callback can expect to be invoked with the
            deserialized response message.
        on_error_callback (:obj:`Callable[[Exception], None]`):
            Callback invoked when an unhandled error occurs at the protocol
            level.
    """

    TIMER_TASK_DELAY = int(os.getenv("PUB_UMB_TIMER_TASK_DELAY", "10"))
    """int: The default delay after which scheduled timer tasks should run."""

    _PROTON_KNOWN_UNHANDLED_ERRORS = frozenset(
        ("amqp:resource-limit-exceeded#local-idle-timeout expired",)
    )
    """frozenset: A set of {error.name}#{error.description} tags.

    There are errors that Proton handles itself using built-in Handler
    implementations e.g. start attempting to reconnect when a connection
    error occurs.

    These are errors that Proton has demonstrated it does not handle
    e.g. local-idle-timeout from RHELDST-8317.
    """

    def __init__(
        self,
        settings: UMBSettings,
        claim_messages: List[Dict[str, Any]],
        message_sender_callback: Callable[[Any], Any],
        on_message_callback: Callable[[Any], None] = do_nothing,
        on_error_callback: Callable[[Any], None] = raise_error,
    ) -> None:
        super(ManifestClaimsHandler, self).__init__()
        self.umb_urls = settings.broker_urls
        self.radas_address = settings.radas_address
        self.claim_messages = list(claim_messages)
        self.timeout = settings.signing_timeout
        self.throttle = settings.signing_throttle
        self.retry = settings.signing_retry
        self.timer_task = None
        self.receiver = None
        self.message_sender_callback = message_sender_callback
        self._on_next = on_message_callback
        self._on_error = on_error_callback
        self.awaiting_response: dict[str, Any] = {}  # {request_id: monotonic.monotonic()}
        self.retry_count: dict[str, int] = {}  # {request_id: 1/2}
        # a mutable list caches messages to send
        self.to_send = list(claim_messages)
        # {request_id: message} a map used to find wanted message by request_id
        self.id_msg_map = {}
        for msg in self.claim_messages:
            self.id_msg_map[msg["request_id"]] = msg  # pragma: no cover
        self.connected = False

        self.ssl_domain = proton.SSLDomain(proton.SSLDomain.MODE_CLIENT)
        self.ssl_domain.set_credentials(settings.pub_cert, settings.pub_cert, None)
        self.ssl_domain.set_trusted_ca_db(settings.ca_cert)
        self.ssl_domain.set_peer_authentication(proton.SSLDomain.ANONYMOUS_PEER)

    def on_start(self, event: proton.Event) -> None:  # pragma: no cover
        LOG.debug("Message event loop starting, connecting to brokers...")
        conn = event.container.connect(
            urls=self.umb_urls, ssl_domain=self.ssl_domain, sasl_enabled=False
        )
        self.receiver = event.container.create_receiver(conn, self.radas_address)
        self.timer_task = event.container.schedule(self.timeout, self)
        # schedule a timer task, if the connection to UMB could be established, raise exception.
        LOG.debug("Message event loop started")

    def on_timer_task(self, event: proton.Event) -> None:
        """timer task has three functionalities:
        1. if it couldn't be connected to brokers, then after self.timeout, exception
           will be raised.
        2. Check if any request is timed out, if it is, then continue checking the retry history
            - if the retry history shows it's been tried more than 3 times, then stop loop.
            - if not, then remove it from awaiting_response dict and send it back to to_send
              queue, bump the retry count.
        3. Check if the number of requests in processing has reached throttle or not, if not
           then send corresponding number of messages.
        """
        if not self.connected:
            LOG.error("Couldn't connect to brokers after %s seconds", self.timeout)
            event.container.stop()
            raise MessageHandlerTimeoutException()
        for request_id, started in list(self.awaiting_response.items()):
            if monotonic.monotonic() - started > self.timeout:
                if request_id not in self.retry_count or self.retry_count[request_id] < self.retry:
                    self.retry_count.setdefault(request_id, 0)
                    self.retry_count[request_id] += 1
                    LOG.warning(
                        "Didn't receive response in %s for request %s, will retry [%s/%s]",
                        self.timeout,
                        request_id,
                        self.retry_count[request_id],
                        self.retry,
                    )
                    # append to resend queue and remove from awaiting_response queue
                    self.to_send.append(self.id_msg_map[request_id])
                    self.awaiting_response.pop(request_id)
                else:
                    LOG.warning("Stopping message event loop due to timeout %s", request_id)
                    event.container.stop()
                    raise MessageHandlerTimeoutException()

        # send more requests if number of waiting < throttle
        spots = max(0, self.throttle - len(self.awaiting_response))
        if spots:
            # if there's free spots, send messages in queue
            self._send_message(min(len(self.to_send), spots))

        # schedule the next timer task
        self.timer_task = event.container.schedule(self.TIMER_TASK_DELAY, self)

    def on_link_opened(self, event: proton.Event) -> None:
        # do the message sending to when the connection has been opened
        # as there is the potential for messages to be dropped.
        if event.receiver == self.receiver:  # pragma: no cover
            # link's open, cancel the scheduled timer task, which was for connecting
            # timing out.
            self.timer_task.cancel()  # type: ignore
            self.connected = True
            self._send_message()
            self.timer_task = event.container.schedule(self.TIMER_TASK_DELAY, self)
        else:
            LOG.warning("Unexpected on_link_opened event")

    def on_message(self, event: proton.Event) -> None:
        """Once receive a message, check if it's expected by checking the awaiting_response,
        if it is, then append it to received and remove it from awaiting_response.
        """
        outer_message = json.loads(event.message.body)
        radas_message = outer_message["msg"]
        request_id = radas_message["request_id"]

        if request_id in self.awaiting_response:  # pragma: no cover
            LOG.info("Received signing response: %s", request_id)
            self.awaiting_response.pop(request_id)
            self._on_next(radas_message)
            if not self.to_send and not self.awaiting_response:
                LOG.info("All requests satisfied, closing connection...")
                self.receiver.close()  # type: ignore
                event.connection.close()
                self.timer_task.cancel()  # type: ignore
                LOG.info("Connection closed.")
        else:
            LOG.debug("Ignored signing response: %s", radas_message["request_id"])

    def on_connection_closing(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: connection_closing")

    def on_connection_closed(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: connection_closed")

    def on_connection_error(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: connection_error")
        self._endpoint_error(event, event.connection)

    def on_session_closing(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: session_closing")

    def on_session_closed(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: session_closed")

    def on_session_error(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: session_error")
        self._endpoint_error(event, event.session)

    def on_link_closing(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: link_closing")

    def on_link_closed(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: link_closed")

    def on_link_error(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: link_error")
        self._endpoint_error(event, event.link)

    def on_transport_error(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: transport_error")
        # transport errors appear to be handled by Proton
        # in most cases, so check for a known case where
        # that isn't true, and consider that an actual error
        self._endpoint_error_if_unhandled(event, event.transport)

    def on_disconnected(self, event: proton.Event) -> None:
        LOG.debug("Messaging event: disconnected")

    def _endpoint_error(self, event: proton.Event, endpoint: str) -> None:
        condition = _get_endpoint_error_condition(endpoint)
        if condition is None:
            error = AMQPEndpointError(endpoint_name=endpoint.__class__.__name__)
        else:
            error = AMQPEndpointError(
                endpoint_name=endpoint.__class__.__name__,
                error_label=condition.name,
                description=condition.description,
            )
        LOG.error(str(error))
        event.container.stop()
        self._on_error(error)

    def _endpoint_error_if_unhandled(self, event: proton.Event, endpoint: proton.Endpoint) -> None:
        condition = _get_endpoint_error_condition(endpoint)
        if condition is not None:
            error_tag = "{condition.name}#{condition.description}".format(condition=condition)
            if error_tag in self._PROTON_KNOWN_UNHANDLED_ERRORS:
                self._endpoint_error(event, endpoint)

    def _send_message(self, count: Optional[int] = None) -> None:
        if not self.to_send:  # pragma: no cover
            return

        if count is None:  # pragma: no cover
            count = min(len(self.to_send), self.throttle)

        messages = self.to_send[:count]
        LOG.info("Sending %s messages...", count)
        self.message_sender_callback(messages)
        # remove sent message from sending queue
        del self.to_send[:count]

        # add sent message's request id to waiting queue
        for msg in messages:
            self.awaiting_response[msg["request_id"]] = monotonic.monotonic()


class _ManifestClaimsRunner(object):
    """Wrap the execution of ManifestClaimsHandler and manage a bit of its lifecycle.

    This class handles restarting messaging in the event of an error, and
    imposes a maximum retry limit on the number of times messaging will
    be restarted.

    Parameters:
        settings (:obj:`UMBSettings`): A UMBSettings object prescribing the
            correct settings for the remote UMB peer through which communication
            with Radas happens.
        claim_messages (:obj:`Sequence[Mapping[str, Any]]`): The manifest claim
            messages that need to be signed by Radas.
        send_action (:obj:`Callable[[Sequence[Mapping[str, Any]]], None]:
            A callback that will be invoked with to send messages to Radas for
            signing.

    See Also:
        `ManifestClaimsHandler`

    """

    def __init__(
        self,
        settings: UMBSettings,
        claim_messages: List[Dict[str, Any]],
        send_action: Callable[[Sequence[Mapping[str, Any]]], None],
    ):
        self._settings = settings
        self._claim_messages = claim_messages
        self._send_action = send_action
        self._retry_attempts = 0
        self._maximum_retries = settings.signing_retry
        self._received_messages: dict[Any, Any] = {}

    @property
    def received_messages(self) -> Sequence[Mapping[str, Any]]:  # pragma: no cover
        """:obj:`Sequence[Mapping[str, Any]]`: The set of received messages."""
        return list(self._received_messages.values())

    def on_next(self, message: Dict[str, Any]) -> None:  # pragma: no cover
        """Track incoming response messages."""
        request_id = message["request_id"]
        self._received_messages.setdefault(request_id, message)

    def on_error(self, error: str) -> None:
        """Determines if messaging can and needs to be restarted, and does so if necessary."""
        LOG.error("Error in message handler: %s", error)

        if self._retry_attempts > self._maximum_retries:
            raise RuntimeError(
                f"Retry limit reached. Message handler has failed with an error: {error}"
            )

        missing = [
            msg for msg in self._claim_messages if msg["request_id"] not in self._received_messages
        ]

        if not missing:
            LOG.info("No claims missing responses. Restart not required.")
            return

        self._retry_attempts += 1
        LOG.info(
            "Restarting message handler for %d remaining claims (retry %d/%d).",
            len(missing),
            self._retry_attempts,
            self._maximum_retries,
        )
        self._run(missing)

    def start(self) -> None:  # pragma: no cover
        """Start manifest claim messaging for the first time."""
        self._run(self._claim_messages)

    def _run(self, claims: List[Dict[str, Any]]) -> None:  # pragma: no cover
        """Run manifest claim messaging with the given set of claims."""
        handler = ManifestClaimsHandler(
            self._settings,
            claims,
            self._send_action,
            on_message_callback=self.on_next,
            on_error_callback=self.on_error,
        )
        Container(handler).run()



import json
import logging
import mock
import pytest
import requests_mock
from copy import deepcopy

from pubtools._quay.push_docker import PushDocker
from pubtools._quay.tag_docker import TagDocker
from pubtools._quay import iib_operations
from pubtools._quay import clear_repo
from pubtools._quay import remove_repo
from pubtools._quay.exceptions import BadPushItem
from .utils.misc import (
    sort_dictionary_sortable_values,
    compare_logs,
    IIBRes,
    mock_manifest_list_requests,
)

# flake8: noqa: E501

MSG_SIGNER_OPERATION_RESULT = [
    [
        {
            "i": 0,
            "msg": {
                "errors": [],
                "manifest_digest": "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6",
                "pub_task_id": "12345",
                "repo": "iib-namespace/new-index-image",
                "request_id": "89cf86e0-8403-46e0-b5ed-5984a635e89e",
                "request_received_time": "2023-10-17T08:08:01.544757",
                "sig_key_id": "37036783",
                "sig_keyname": "testing",
                "signature_type": "container_signature",
                "signed_claim": "claim1",
            },
        },
        {},
    ]
]


@mock.patch("pubtools._quay.push_docker.SecurityManifestPusher")
@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.utils.misc.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
@mock.patch("pubtools._quay.push_docker.timestamp")
def test_push_docker_multiarch_merge_ml_operator(
    mock_timestamp,
    mock_run_cmd,
    mock_run_entrypoint_operator_pusher,
    mock_run_entrypoint_misc,
    mock_api_client,
    mock_security_manifest_pusher,
    target_settings,
    container_multiarch_push_item_integration,
    operator_push_item_ok,
    src_manifest_list,
    dest_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    signer_wrapper_remove_signatures,
    fixture_run_in_parallel_signer,
    fixture_run_in_parallel_push_docker,
):
    mock_timestamp.return_value = "timestamp"
    # hub usage has to be mocked
    hub = mock.MagicMock()
    mock_get_target_info = mock.MagicMock()
    mock_get_target_info.return_value = {
        "settings": {
            "quay_namespace": "stage-namespace",
            "dest_quay_user": "stage-user",
            "dest_quay_password": "stage-password",
            "pyxis_server": "",
        }
    }
    hub.worker.get_target_info = mock_get_target_info
    target_settings["propagated_from"] = "test-target"

    mock_run_entrypoint_misc.side_effect = [
        # pubtools-pyxis-get-repo-metadata
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        # pubtools-pyxis-get-repo-metadata in get_index_images
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]
    signer_wrapper_run_entry_point.side_effect = [
        # fetch existing signatures
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:5555555555x",
                "reference": "some-registry1.com/target/repo:latest-test-tag",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        # store signed
        [],
        [],
        # fetch existing manfiests
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666a",
                "reference": "some-registry1.com/namespace/operators/index-image:v4.5",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666a",
                "reference": "some-registry1.com/namespace/operators/index-image:v4.5",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        # store signed
        [],
        # filter existing manifests
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666",
                "reference": "some-registry1.com/namespace/operators/index-image:v4.5",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        # list existing signatures
        (True, ["quay.io/testing/repo:v4.5"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig1"]),
        (True, ["quay.io/testing/repo:v4.6"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig1"]),
        [],
        [],
        [],
        # filter to remove
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666",
                "reference": "some-registry1.com/operators/index-image:v4.5",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        # filter to remove
        [],
        [],
        [],
        [],
        [],
        [],
        (True, ["quay.io/testing/repo:v4.5"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig1"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig1"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig2"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig3"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig4"]),
        # remove existing signatures
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666",
                "reference": "some-registry2.com/operators/index-image:v4.5",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        [
            {
                "_id": 1,
                "manifest_digest": "sha256:6666666666",
                "reference": "some-registry1.com/operators/index-image:v4.6",
                "sig_key_id": "some-key",
                "repository": "operators/index-image",
            }
        ],
        # remove signatures (list signatures)
        (True, ["quay.io/testing/repo:sha256-6666666666.sig5"]),
        (True, ["quay.io/testing/repo:sha256-6666666666.sig6"]),
    ]
    mock_run_entrypoint_operator_pusher.side_effect = [
        # pubtools-pyxis-get-operator-indices
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
        # pubtools-iib-add-bundles (4.5)
        IIBRes(
            "registry.com/namespace/index-image@sha256:v4.5",
            "registry.com/namespace/iib@sha256:a1a1a1",
            ["v4.5-1"],
        ),
        # pubtools-iib-add-bundles (4.6)
        IIBRes(
            "registry.com/namespace/index-image@sha256:v4.6",
            "registry.com/namespace/iib@sha256:b2b2b2",
            ["v4.6-1"],
        ),
    ]

    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/stage-namespace/test_namespace----test_repo/tags/list",
            json={"some-data": "value"},
        )
        mock_manifest_list_requests(
            m, "https://quay.io/v2/src/repo/manifests/1", src_manifest_list, v2s1_manifest
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            dest_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/sha256:9daac465523ce42a89e605151734e7b92c5ade2123055a6a2aeabbf60e5edfa4",
            dest_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/v4.5",
            dest_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/v4.6",
            dest_manifest_list,
            v2s1_manifest,
        )
        m.put("https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag")
        m.get("https://git-server.com/v4_5.yml/raw?ref=master")
        m.get("https://git-server.com/v4_6.yml/raw?ref=master")
        mock_manifest_list_requests(
            m, "https://quay.io/v2/namespace/iib/manifests/v4.5-1", src_manifest_list, v2s1_manifest
        )
        mock_manifest_list_requests(
            m, "https://quay.io/v2/namespace/iib/manifests/v4.6-1", src_manifest_list, v2s1_manifest
        )
        m.delete("https://pyxis-url.com/v1/signatures/id/some-id1")
        m.delete("https://pyxis-url.com/v1/signatures/id/some-id2")
        mock_manifest_list_requests(
            m, "https://some-registry1.com/v2/repo/manifests/1.0", src_manifest_list, v2s1_manifest
        )
        signer_wrapper_entry_point.return_value = {
            "signer_result": {
                "status": "ok",
            },
            "operation": {
                "references": ["some-registry.com/iib-namespace/new-index-image:8"],
                "manifests": [
                    "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
                ],
            },
            "operation_results": MSG_SIGNER_OPERATION_RESULT,
            "signing_key": "sig-key",
        }

        push_docker = PushDocker(
            [container_multiarch_push_item_integration, operator_push_item_ok],
            hub,
            "1",
            "some-target",
            target_settings,
        )

        push_docker.run()
        signer_wrapper_entry_point.assert_has_calls(
            [
                # msg signing wrapper
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                    digest=[
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                # cosign
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                    ],
                    digest=[
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "sha256:1111111111",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/operators/index-image:v4.5",
                        "some-registry1.com/operators/index-image:v4.5-timestamp",
                        "some-registry2.com/operators/index-image:v4.5",
                        "some-registry2.com/operators/index-image:v4.5-timestamp",
                    ],
                    digest=[
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/operators/index-image:v4.6",
                        "some-registry1.com/operators/index-image:v4.6-timestamp",
                        "some-registry2.com/operators/index-image:v4.6",
                        "some-registry2.com/operators/index-image:v4.6-timestamp",
                    ],
                    digest=[
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/operators----index-image:v4.5",
                        "quay.io/some-namespace/operators----index-image:v4.5",
                        "quay.io/some-namespace/operators----index-image:v4.5-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.5-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.5",
                        "quay.io/some-namespace/operators----index-image:v4.5-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.5",
                        "quay.io/some-namespace/operators----index-image:v4.5-timestamp",
                    ],
                    digest=[
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/operators/index-image:v4.5",
                        "some-registry2.com/operators/index-image:v4.5",
                        "some-registry1.com/operators/index-image:v4.5-timestamp",
                        "some-registry2.com/operators/index-image:v4.5-timestamp",
                        "some-registry1.com/operators/index-image:v4.5",
                        "some-registry1.com/operators/index-image:v4.5-timestamp",
                        "some-registry2.com/operators/index-image:v4.5",
                        "some-registry2.com/operators/index-image:v4.5-timestamp",
                    ],
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/operators----index-image:v4.6",
                        "quay.io/some-namespace/operators----index-image:v4.6",
                        "quay.io/some-namespace/operators----index-image:v4.6-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.6-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.6",
                        "quay.io/some-namespace/operators----index-image:v4.6-timestamp",
                        "quay.io/some-namespace/operators----index-image:v4.6",
                        "quay.io/some-namespace/operators----index-image:v4.6-timestamp",
                    ],
                    digest=[
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/operators/index-image:v4.6",
                        "some-registry2.com/operators/index-image:v4.6",
                        "some-registry1.com/operators/index-image:v4.6-timestamp",
                        "some-registry2.com/operators/index-image:v4.6-timestamp",
                        "some-registry1.com/operators/index-image:v4.6",
                        "some-registry1.com/operators/index-image:v4.6-timestamp",
                        "some-registry2.com/operators/index-image:v4.6",
                        "some-registry2.com/operators/index-image:v4.6-timestamp",
                    ],
                ),
            ]
        )
        assert signer_wrapper_remove_signatures.mock_calls == [
            mock.call([1]),
            # TODO: uncomment when removing of signatures in cosign is enabled
            # mock.call([("operators/index-image", "sha256:6666666666")]),
        ]


@mock.patch("pubtools._quay.push_docker.SecurityManifestPusher")
@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
def test_push_docker_multiarch_simple_workflow(
    mock_run_cmd,
    mock_api_client,
    mock_security_manifest_pusher,
    target_settings,
    container_multiarch_push_item_integration,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    fixture_run_in_parallel,
):
    # hub usage has to be mocked
    hub = mock.MagicMock()
    mock_get_target_info = mock.MagicMock()
    mock_get_target_info.return_value = {
        "settings": {
            "quay_namespace": "stage-namespace",
            "dest_quay_user": "stage-user",
            "dest_quay_password": "stage-password",
        }
    }
    hub.worker.get_target_info = mock_get_target_info
    target_settings["propagated_from"] = "test-target"
    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/stage-namespace/test_namespace----test_repo/tags/list",
            json={"some-data": "value"},
        )
        mock_manifest_list_requests(
            m, "https://quay.io/v2/src/repo/manifests/1", src_manifest_list, v2s1_manifest
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/sha256:8ce181d89b7bb7f1639d8df3d65d630b1322d0bb6daff5c492eec24ec53628d5",
            src_manifest_list,
            v2s1_manifest,
        )
        m.put(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
        )

        signer_wrapper_entry_point.return_value = {
            "signer_result": {
                "status": "ok",
            },
            "operation": {
                "references": ["some-registry.com/iib-namespace/new-index-image:8"],
                "manifests": [
                    "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
                ],
            },
            "operation_results": MSG_SIGNER_OPERATION_RESULT,
            "signing_key": "sig-key",
        }

        push_docker = PushDocker(
            [container_multiarch_push_item_integration],
            hub,
            "1",
            "some-target",
            target_settings,
        )

        push_docker.run()
        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                    digest=[
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                # cosign
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                    ],
                    digest=[
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "sha256:1111111111",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                ),
            ]
        )


@mock.patch("pubtools._quay.push_docker.SecurityManifestPusher")
@mock.patch("pubtools._quay.push_docker.PushDocker.fetch_missing_push_items_digests")
@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
def test_push_docker_source(
    mock_run_cmd,
    mock_api_client,
    mock_fetch_missing_push_items_digests,
    mock_security_manifest_pusher,
    target_settings,
    container_source_push_item_integration,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    # hub usage has to be mocked
    hub = mock.MagicMock()
    mock_get_target_info = mock.MagicMock()
    mock_get_target_info.return_value = {
        "settings": {
            "quay_namespace": "stage-namespace",
            "dest_quay_user": "stage-user",
            "dest_quay_password": "stage-password",
        }
    }
    hub.worker.get_target_info = mock_get_target_info
    target_settings["propagated_from"] = "test-target"

    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}

    with requests_mock.Mocker() as m:
        mock_manifest_list_requests(
            m, "https://quay.io/v2/src/repo/manifests/1", src_manifest_list, v2s1_manifest
        )
        m.get(
            "https://quay.io/v2/stage-namespace/test_namespace----test_repo/tags/list",
            json={"some-data": "value"},
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            src_manifest_list,
            v2s1_manifest,
        )
        m.get(
            "https://quay.io/v2/some-namespace/target----repo/manifests/1.0",
            text=json.dumps(v2s1_manifest, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            text=json.dumps(v2s1_manifest, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v1+json"},
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/sha256:8ce181d89b7bb7f1639d8df3d65d630b1322d0bb6daff5c492eec24ec53628d5",
            src_manifest_list,
            v2s1_manifest,
        )

        m.put(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
        )

        signer_wrapper_entry_point.return_value = {
            "signer_result": {
                "status": "ok",
            },
            "operation": {
                "references": ["some-registry.com/iib-namespace/new-index-image:8"],
                "manifests": [
                    "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
                ],
            },
            "operation_results": MSG_SIGNER_OPERATION_RESULT,
            "signing_key": "sig-key",
        }

        push_docker = PushDocker(
            [container_source_push_item_integration],
            hub,
            "1",
            "some-target",
            target_settings,
        )

        def mock_fetch_missing_push_items_digests_sf(push_items):
            fake_digest_counter = 0
            ret = {}
            for item in push_items:
                ret["quay.io"] = {
                    "target/repo": {
                        "latest-test-tag": {
                            "application/vnd.docker.distribution.manifest.v2+json": (
                                "fake-digest-%s" % fake_digest_counter,
                                "fake-sign-key",
                            )
                        }
                    }
                }
                fake_digest_counter += 1
            return ret

        mock_fetch_missing_push_items_digests.side_effect = mock_fetch_missing_push_items_digests_sf

        push_docker.run()
        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="fake-sign-key",
                    reference=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                    digest=["fake-digest-0", "fake-digest-0"],
                    task_id="1",
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="fake-sign-key",
                    reference=[
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                    ],
                    digest=["fake-digest-0", "fake-digest-0"],
                    identity=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                ),
            ]
        )


@mock.patch("subprocess.Popen")
def test_tag_docker_multiarch_merge_ml(
    mock_popen,
    target_settings,
    tag_docker_push_item_add_integration,
    tag_docker_push_item_remove_no_src_integration,
    src_manifest_list,
    fake_cert_key_paths,
    v2s1_manifest,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    hub = mock.MagicMock()
    mock_get_target_info = mock.MagicMock()
    mock_get_target_info.return_value = {
        "settings": {
            "quay_namespace": "stage-namespace",
            "dest_quay_user": "stage-user",
            "dest_quay_password": "stage-password",
        }
    }
    hub.worker.get_target_info = mock_get_target_info
    target_settings["propagated_from"] = "test-target"

    mock_popen.return_value.communicate.return_value = ("Login Succeeded", "err")
    mock_popen.return_value.returncode = 0

    src_manifest_list_missing = deepcopy(src_manifest_list)
    src_manifest_list_missing["manifests"] = src_manifest_list_missing["manifests"][:2]

    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }

    with requests_mock.Mocker() as m:
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.5",
            src_manifest_list,
            v2s1_manifest,
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo/tags/list",
            json={
                "name": "namespace----test_repo",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/tags/list",
            json={
                "name": "namespace----test_repo",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo2/tags/list",
            json={
                "name": "namespace----test_repo2",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/tags/list",
            json={
                "name": "namespace----test_repo2",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo/manifests/v1.6",
            json=src_manifest_list_missing,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo2/manifests/v1.8",
            status_code=404,
        )
        m.get(
            "https://quay.io/api/v1/repository/some-namespace/namespace----test_repo?includeTags=True",
            json={
                "tags": {
                    "v1.5": {"manifest_digest": "a1a1a1"},
                    "v1.6": {"manifest_digest": "b2b2b2"},
                }
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.6",
            status_code=404,
            text="Not Found",
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.put(
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.6",
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.8",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/api/v1/repository/some-namespace/namespace----test_repo2?includeTags=True",
            json={"tags": {"v1.8": {"manifest_digest": "c3c3c3"}}},
        )
        m.put(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.8",
        )
        # manifests for removal of old signatures
        m.get(
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/manifest_list_digest",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/sha256:1111111111",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/sha256:5555555555",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )

        tag_docker_instance = TagDocker(
            [tag_docker_push_item_add_integration, tag_docker_push_item_remove_no_src_integration],
            hub,
            "1",
            "some-target",
            target_settings,
        )
        tag_docker_instance.run()
        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                    ],
                    digest=[
                        "sha256:1111111111",
                        "sha256:1111111111",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                # cosign
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                    ],
                    digest=[
                        "sha256:1111111111",
                        "sha256:1111111111",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:71e75d5344d529631eaf40a8f9522edb7a66620d73eda6aff667572d511c6519",
                        "sha256:71e75d5344d529631eaf40a8f9522edb7a66620d73eda6aff667572d511c6519",
                    ],
                    identity=[
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                    ],
                ),
            ]
        )


@mock.patch("subprocess.Popen")
@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
def test_tag_docker_source_copy_untag(
    mock_run_cmd,
    mock_api_client,
    mock_popen,
    target_settings,
    tag_docker_push_item_add_integration,
    tag_docker_push_item_remove_no_src_integration,
    v2s2_manifest_data,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    hub = mock.MagicMock()
    mock_get_target_info = mock.MagicMock()
    mock_get_target_info.return_value = {
        "settings": {
            "quay_namespace": "stage-namespace",
            "dest_quay_user": "stage-user",
            "dest_quay_password": "stage-password",
            "signing": [{"enabled": True, "label": "msg_signer", "config_file": "test-config.yml"}],
        }
    }
    hub.worker.get_target_info = mock_get_target_info
    target_settings["propagated_from"] = "test-target"

    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}
    mock_popen.return_value.communicate.side_effect = [
        ("something", "err"),
        ("Login Succeeded", "err"),
        ('{"Architecture": "amd64"}', "err"),
        ('{"Architecture": "amd64"}', "err"),
        ('{"Architecture": "amd64"}', "err"),
        ('{"Architecture": "amd64"}', "err"),
    ]
    mock_popen.return_value.returncode = 0
    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.5",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.5",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/test_repo2/manifests/v1.6",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo/tags/list",
            json={
                "name": "namespace----test_repo",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/tags/list",
            json={
                "name": "namespace----test_repo",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo2/tags/list",
            json={
                "name": "namespace----test_repo2",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/tags/list",
            json={
                "name": "namespace----test_repo2",
                "tags": [
                    "v1.5",
                    "v1.6",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo/manifests/v1.6",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.6",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/stage-namespace/namespace----test_repo2/manifests/v1.8",
            status_code=404,
        )
        m.get(
            "https://quay.io/api/v1/repository/some-namespace/test_repo?includeTags=True",
            json={
                "tags": {
                    "v1.5": {"manifest_digest": "a1a1a1"},
                    "v1.6": {"manifest_digest": "b2b2b2"},
                }
            },
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.6",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.put(
            "https://quay.io/v2/some-namespace/test_repo/manifests/v1.6",
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.8",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/api/v1/repository/some-namespace/test_repo2?includeTags=True",
            json={"tags": {"v1.8": {"manifest_digest": "c3c3c3", "image_id": "some-id"}}},
        )
        m.put(
            "https://quay.io/v2/some-namespace/test_repo2/manifests/v1.8",
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-namespace/namespace----test_repo2/tag/v1.8",
        )

        # manifest for removal of old signatures
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
            json=v2s2_manifest_data,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        tag_docker_instance = TagDocker(
            [tag_docker_push_item_add_integration, tag_docker_push_item_remove_no_src_integration],
            hub,
            "1",
            "some-target",
            target_settings,
        )

        tag_docker_instance.run()
        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                    ],
                    digest=[
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                    ],
                    task_id="1",
                ),
                # cosign
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                        "quay.io/some-namespace/namespace----test_repo:v1.6",
                    ],
                    digest=[
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                    ],
                    identity=[
                        "some-registry1.com/namespace/test_repo:v1.6",
                        "some-registry2.com/namespace/test_repo:v1.6",
                    ],
                ),
            ]
        )


# @mock.patch("subprocess.Popen")
# @mock.patch("pubtools._quay.command_executor.docker.APIClient")
# @mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
# def test_tag_docker_source_copy_untag_missing_source_tags(
#     mock_run_cmd,
#     mock_api_client,
#     mock_popen,
#     target_settings,
#     tag_docker_push_item_add_integration,
#     tag_docker_push_item_remove_no_src_integration,
#     v2s2_manifest_data,
#     fake_cert_key_paths,
#     signer_wrapper_entry_point,
#     signer_wrapper_run_entry_point,
# ):
#     hub = mock.MagicMock()
#     mock_get_target_info = mock.MagicMock()
#     mock_get_target_info.return_value = {
#         "settings": {
#             "quay_namespace": "stage-namespace",
#             "dest_quay_user": "stage-user",
#             "dest_quay_password": "stage-password",
#             "signing": [{"enabled": True, "label": "msg_signer", "config_file": "test-config.yml"}],
#         }
#     }
#     hub.worker.get_target_info = mock_get_target_info
#     target_settings["propagated_from"] = "test-target"
#
#     mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
#     mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}
#     mock_popen.return_value.communicate.side_effect = [
#         ("something", "err"),
#         ("Login Succeeded", "err"),
#         ('{"Architecture": "amd64"}', "err"),
#         ('{"Architecture": "amd64"}', "err"),
#         ('{"Architecture": "amd64"}', "err"),
#         ('{"Architecture": "amd64"}', "err"),
#     ]
#     mock_popen.return_value.returncode = 0
#     signer_wrapper_entry_point.return_value = {
#         "signer_result": {
#             "status": "ok",
#         },
#         "operation": {
#             "references": ["some-registry.com/iib-namespace/new-index-image:8"],
#             "manifests": [
#                 "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
#             ],
#         },
#         "operation_results": MSG_SIGNER_OPERATION_RESULT,
#         "signing_key": "sig-key",
#     }
#
#     with requests_mock.Mocker() as m:
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.5",
#             status_code=404,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.5",
#             json=v2s2_manifest_data,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/test_repo2/manifests/v1.6",
#             json=v2s2_manifest_data,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/v2/stage-namespace/namespace----test_repo/tags/list",
#             json={
#                 "name": "namespace----test_repo",
#                 "tags": [
#                     "v1.5",
#                     "v1.6",
#                 ],
#             },
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo/tags/list",
#             json={
#                 "name": "namespace----test_repo",
#                 "tags": [
#                     "v1.5",
#                     "v1.6",
#                 ],
#             },
#         )
#         m.get(
#             "https://quay.io/v2/stage-namespace/namespace----test_repo2/tags/list",
#             json={
#                 "name": "namespace----test_repo2",
#                 "tags": [
#                     "v1.5",
#                     "v1.6",
#                 ],
#             },
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo2/tags/list",
#             json={
#                 "name": "namespace----test_repo2",
#                 "tags": [
#                     "v1.5",
#                     "v1.6",
#                 ],
#             },
#         )
#         m.get(
#             "https://quay.io/v2/stage-namespace/namespace----test_repo/manifests/v1.6",
#             json=v2s2_manifest_data,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo/manifests/v1.6",
#             json=v2s2_manifest_data,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/v2/stage-namespace/namespace----test_repo2/manifests/v1.8",
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#             json=v2s2_manifest_data,
#         )
#         m.get(
#             "https://quay.io/api/v1/repository/some-namespace/test_repo?includeTags=True",
#             json={
#                 "tags": {
#                     "v1.5": {"manifest_digest": "a1a1a1"},
#                     "v1.6": {"manifest_digest": "b2b2b2"},
#                 }
#             },
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.6",
#             status_code=404
#         )
#         m.put(
#             "https://quay.io/v2/some-namespace/test_repo/manifests/v1.6",
#         )
#         m.get(
#             "https://quay.io/v2/some-namespace/namespace----test_repo2/manifests/v1.8",
#             json=v2s2_manifest_data,
#             headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
#         )
#         m.get(
#             "https://quay.io/api/v1/repository/some-namespace/test_repo2?includeTags=True",
#             json={"tags": {"v1.8": {"manifest_digest": "c3c3c3", "image_id": "some-id"}}},
#         )
#         m.put(
#             "https://quay.io/v2/some-namespace/test_repo2/manifests/v1.8",
#         )
#         m.delete(
#             "https://quay.io/api/v1/repository/some-namespace/namespace----test_repo2/tag/v1.8",
#         )
#
#         tag_docker_instance = TagDocker(
#             [tag_docker_push_item_add_integration, tag_docker_push_item_remove_no_src_integration],
#             hub,
#             "1",
#             "some-target",
#             target_settings,
#         )
#         with pytest.raises(BadPushItem):
#             tag_docker_instance.run()


@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.iib_operations.timestamp")
def test_task_iib_add_bundles(
    mock_timestamp,
    mock_run_entrypoint_operator_pusher,
    mock_run_cmd,
    mock_api_client,
    target_settings,
    src_manifest_list,
    fake_cert_key_paths,
    v2s1_manifest,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    fixture_run_in_parallel,
):
    mock_timestamp.return_value = "timestamp"
    build_details = IIBRes(
        "some-registry.com/iib-namespace/new-index-image:8",
        "some-registry.com/iib-namespace/iib@sha256:a1a1a1",
        ["8-1"],
    )
    mock_run_entrypoint_operator_pusher.return_value = build_details
    mock_run_cmd.side_effect = [("Login Succeeded", "err"), ("Login Succeeded", "err")]
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}

    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }

    with requests_mock.Mocker() as m:
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/iib-namespace/iib/manifests/8-1",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/8",
            src_manifest_list,
            v2s1_manifest,
        )
        m.get(
            "https://quay.io/v2/iib-namespace/iib/manifests/sha256:a1a1a1",
            json=src_manifest_list,
            headers={
                "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json",
                "docker-content-digest": "sha256:a1a1a1",
            },
        )
        # manifest to remove old signatures
        m.get(
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/manifest_list_digest",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )

        iib_operations.task_iib_add_bundles(
            ["bundle1", "bundle2"],
            ["arch1", "arch2"],
            "some-registry.com/redhat-namespace/new-index-image:5",
            ["bundle3", "bundle4"],
            ["some-key"],
            "1",
            target_settings,
        )
        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/operators/index-image:8",
                        "some-registry1.com/operators/index-image:8-timestamp",
                        "some-registry2.com/operators/index-image:8",
                        "some-registry2.com/operators/index-image:8-timestamp",
                    ],
                    digest=[
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                # cosign
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/operators----index-image:8",
                        "quay.io/some-namespace/operators----index-image:8",
                        "quay.io/some-namespace/operators----index-image:8-timestamp",
                        "quay.io/some-namespace/operators----index-image:8-timestamp",
                        "quay.io/some-namespace/operators----index-image:8",
                        "quay.io/some-namespace/operators----index-image:8-timestamp",
                        "quay.io/some-namespace/operators----index-image:8",
                        "quay.io/some-namespace/operators----index-image:8-timestamp",
                    ],
                    digest=[
                        "sha256:a1a1a1",
                        "sha256:a1a1a1",
                        "sha256:a1a1a1",
                        "sha256:a1a1a1",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/operators/index-image:8",
                        "some-registry2.com/operators/index-image:8",
                        "some-registry1.com/operators/index-image:8-timestamp",
                        "some-registry2.com/operators/index-image:8-timestamp",
                        "some-registry1.com/operators/index-image:8",
                        "some-registry1.com/operators/index-image:8-timestamp",
                        "some-registry2.com/operators/index-image:8",
                        "some-registry2.com/operators/index-image:8-timestamp",
                    ],
                ),
            ]
        )


@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.iib_operations.timestamp")
def test_task_iib_remove_operators(
    mock_timestamp,
    mock_run_entrypoint_operator_pusher,
    mock_run_cmd,
    mock_api_client,
    target_settings,
    src_manifest_list,
    fake_cert_key_paths,
    v2s1_manifest,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    mock_timestamp.return_value = "timestamp"
    build_details = IIBRes(
        "some-registry.com/iib-namespace/new-index-image:8",
        "some-registry.com/iib-namespace/iib@sha256:a1a1a1",
        ["8-1"],
    )
    mock_run_entrypoint_operator_pusher.return_value = build_details
    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}

    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }

    with requests_mock.Mocker() as m:
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/iib-namespace/iib/manifests/8-1",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/8",
            src_manifest_list,
            v2s1_manifest,
        )
        m.get(
            "https://quay.io/v2/iib-namespace/iib/manifests/sha256:a1a1a1",
            json=src_manifest_list,
            headers={
                "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json",
                "docker-content-digest": "sha256:a1a1a1",
            },
        )
        # manifests for removal of old signatures
        m.get(
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/manifest_list_digest",
            json=src_manifest_list,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        iib_operations.task_iib_remove_operators(
            ["operator1", "operator2"],
            ["arch1", "arch2"],
            "some-registry.com/redhat-namespace/new-index-image:5",
            ["some-key"],
            "1",
            target_settings,
        )

    signer_wrapper_entry_point.assert_has_calls(
        [
            mock.call(
                config_file="test-config.yml",
                signing_key="some-key",
                reference=[
                    "some-registry1.com/operators/index-image:8",
                    "some-registry1.com/operators/index-image:8-timestamp",
                    "some-registry2.com/operators/index-image:8",
                    "some-registry2.com/operators/index-image:8-timestamp",
                ],
                digest=[
                    "sha256:5555555555",
                    "sha256:5555555555",
                    "sha256:5555555555",
                    "sha256:5555555555",
                ],
                task_id="1",
            ),
            # cosign
            mock.call(
                config_file="test-config.yml",
                signing_key="some-key",
                reference=[
                    "quay.io/some-namespace/operators----index-image:8",
                    "quay.io/some-namespace/operators----index-image:8",
                    "quay.io/some-namespace/operators----index-image:8-timestamp",
                    "quay.io/some-namespace/operators----index-image:8-timestamp",
                    "quay.io/some-namespace/operators----index-image:8",
                    "quay.io/some-namespace/operators----index-image:8-timestamp",
                    "quay.io/some-namespace/operators----index-image:8",
                    "quay.io/some-namespace/operators----index-image:8-timestamp",
                ],
                digest=[
                    "sha256:a1a1a1",
                    "sha256:a1a1a1",
                    "sha256:a1a1a1",
                    "sha256:a1a1a1",
                    "sha256:5555555555",
                    "sha256:5555555555",
                    "sha256:5555555555",
                    "sha256:5555555555",
                ],
                identity=[
                    "some-registry1.com/operators/index-image:8",
                    "some-registry2.com/operators/index-image:8",
                    "some-registry1.com/operators/index-image:8-timestamp",
                    "some-registry2.com/operators/index-image:8-timestamp",
                    "some-registry1.com/operators/index-image:8",
                    "some-registry1.com/operators/index-image:8-timestamp",
                    "some-registry2.com/operators/index-image:8",
                    "some-registry2.com/operators/index-image:8-timestamp",
                ],
            ),
        ]
    )


@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_task_iib_build_from_scratch(
    mock_run_entrypoint_operator_pusher,
    mock_run_cmd,
    mock_api_client,
    target_settings,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    build_details = IIBRes(
        "some-registry.com/iib-namespace/new-index-image:12",
        "some-registry.com/iib-namespace/iib@sha256:a1a1a1",
        ["12-1"],
    )
    mock_run_entrypoint_operator_pusher.return_value = build_details
    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}
    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }

    with requests_mock.Mocker() as m:
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/8",
            src_manifest_list,
            v2s1_manifest,
        )
        m.get(
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/manifest_list_digest",
            json=src_manifest_list,
            headers={
                "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json",
                "docker-content-digest": "manifest_list_digest",
            },
        )
        m.get(
            "https://quay.io/v2/iib-namespace/iib/manifests/sha256:a1a1a1",
            text=json.dumps(src_manifest_list, sort_keys=True),
            headers={
                "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json",
                "docker-content-digest": "sha256:a1a1a1",
            },
        )
        iib_operations.task_iib_build_from_scratch(
            ["bundle1", "bundle2"],
            ["arch1", "arch2"],
            "8",
            ["some-key"],
            "1",
            target_settings,
        )


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_clear_repo(
    mock_triangulate,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_remove_signatures,
    signer_wrapper_run_entry_point,
):
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:1111111111",
            "reference": "registry.io/namespace/repo2:3",
            "sig_key_id": "key",
            "repository": "namespace/repo2",
        }
    ]
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/some-org/namespace----repo1/tags/list",
            json={
                "name": "namespace----repo1",
                "tags": [
                    "1",
                    "2",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-org/namespace----repo2/tags/list",
            json={
                "name": "namespace----repo1",
                "tags": [
                    "3",
                ],
            },
        )

        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo1/manifests/1",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo1/manifests/2",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo2/manifests/3",
            src_manifest_list,
            v2s1_manifest,
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-org/namespace----repo1/tag/1",
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-org/namespace----repo1/tag/2",
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-org/namespace----repo2/tag/3",
        )

        clear_repo.clear_repositories(
            repositories="namespace/repo1,namespace/repo2",
            settings=dict(
                quay_org="some-org",
                quay_api_token="some-api-token",
                quay_user="some-user",
                quay_password="some-password",
                pyxis_server="pyxis-server.com",
                pyxis_ssl_crtfile="/path/to/file.crt",
                pyxis_ssl_keyfile="/path/to/file.key",
                pyxis_request_threads=7,
                signers="msg_signer,cosign_signer",
                signer_configs="/test-config.yml,/test-config.yml",
            ),
        )
        signer_wrapper_remove_signatures.assert_called_with([1])


def test_remove_repo(
    src_manifest_list,
    fake_cert_key_paths,
    signer_wrapper_remove_signatures,
    signer_wrapper_run_entry_point,
    v2s1_manifest,
):
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:1111111111",
            "reference": "registry.io/namespace/repo2:3",
            "sig_key_id": "key",
            "repository": "namespace/repo2",
        }
    ]
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/some-org/namespace----repo1/tags/list",
            json={
                "name": "namespace----repo1",
                "tags": [
                    "1",
                    "2",
                ],
            },
        )
        m.get(
            "https://quay.io/v2/some-org/namespace----repo2/tags/list",
            json={
                "name": "namespace----repo1",
                "tags": [
                    "3",
                ],
            },
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo1/manifests/1",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo1/manifests/2",
            src_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-org/namespace----repo2/manifests/3",
            src_manifest_list,
            v2s1_manifest,
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-org/namespace----repo1",
        )
        m.delete(
            "https://quay.io/api/v1/repository/some-org/namespace----repo2",
        )

        remove_repo.remove_repositories(
            repositories="namespace/repo1,namespace/repo2",
            settings=dict(
                quay_org="some-org",
                quay_api_token="some-api-token",
                quay_user="some-user",
                quay_password="some-password",
                pyxis_server="pyxis-server.com",
                pyxis_ssl_crtfile="/path/to/file.crt",
                pyxis_ssl_keyfile="/path/to/file.key",
                pyxis_request_threads=7,
                signers="msg_signer,cosign_signer",
                signer_configs="/test_config.yml,/test_config.yml",
            ),
        )
        signer_wrapper_remove_signatures.assert_called_with([1])


@mock.patch("pubtools._quay.push_docker.SecurityManifestPusher")
@mock.patch("pubtools._quay.command_executor.docker.APIClient")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.command_executor.RemoteExecutor._run_cmd")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_docker_operator_verify_bundle_fail(
    mock_get_repo_metadata,
    mock_run_cmd,
    mock_run_entrypoint_operator_pusher,
    mock_api_client,
    mock_security_manifest_pusher,
    target_settings,
    container_multiarch_push_item_integration,
    operator_push_item_ok,
    src_manifest_list,
    dest_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
    fixture_run_in_parallel_signer,
    fixture_run_in_parallel_push_docker,
    fixture_run_in_parallel_item_processor,
):
    # hub usage has to be mocked
    hub = mock.MagicMock()
    mock_run_entrypoint_operator_pusher.side_effect = [
        # pubtools-pyxis-get-operator-indices
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
    ]
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]
    mock_run_cmd.return_value = ("Login Succeeded", "err")
    mock_api_client.return_value.exec_start.return_value = b"Login Succeeded"
    mock_api_client.return_value.exec_inspect.return_value = {"ExitCode": 0}
    signer_wrapper_entry_point.return_value = {
        "signer_result": {
            "status": "ok",
        },
        "operation": {
            "references": ["some-registry.com/iib-namespace/new-index-image:8"],
            "manifests": [
                "sha256:bd6eba96070efe86b64b9a212680ca6d46a2e30f0a7d8e539f657eabc45c35a6"
            ],
        },
        "operation_results": MSG_SIGNER_OPERATION_RESULT,
        "signing_key": "sig-key",
    }
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/some-namespace/target----repo/tags/list",
            json={"name": "target-repo", "tags": ["latest-test-tag"]},
        )
        mock_manifest_list_requests(
            m, "https://quay.io/v2/src/repo/manifests/1", src_manifest_list, v2s1_manifest
        )
        m.get(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            [
                {
                    "text": json.dumps(dest_manifest_list, sort_keys=True),
                    "headers": {
                        "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"
                    },
                },
                {
                    "json": dest_manifest_list,
                    "headers": {
                        "Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"
                    },
                },
            ],
            request_headers={"Accept": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            text=json.dumps(v2s1_manifest, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/sha256:1111111111",
            text=json.dumps(v2s1_manifest, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
            request_headers={"Accept": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/target----repo/manifests/sha256:9daac465523ce42a89e605151734e7b92c5ade2123055a6a2aeabbf60e5edfa4",
            dest_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/v4.5",
            dest_manifest_list,
            v2s1_manifest,
        )
        mock_manifest_list_requests(
            m,
            "https://quay.io/v2/some-namespace/operators----index-image/manifests/v4.6",
            dest_manifest_list,
            v2s1_manifest,
        )
        m.put(
            "https://quay.io/v2/some-namespace/target----repo/manifests/latest-test-tag",
            text=json.dumps(v2s1_manifest, sort_keys=True),
        )
        m.delete("https://pyxis-url.com/v1/signatures/id/some-id1")
        m.delete("https://pyxis-url.com/v1/signatures/id/some-id2")

        push_docker = PushDocker(
            [container_multiarch_push_item_integration, operator_push_item_ok],
            hub,
            "1",
            "some-target",
            target_settings,
        )

        with pytest.raises(SystemExit, match="1"):
            push_docker.run()

        signer_wrapper_entry_point.assert_has_calls(
            [
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                    digest=[
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:5555555555",
                    ],
                    task_id="1",
                ),
                mock.call(
                    config_file="test-config.yml",
                    signing_key="some-key",
                    reference=[
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                        "quay.io/some-namespace/target----repo:latest-test-tag",
                    ],
                    digest=[
                        "manifest_list_digest",
                        "manifest_list_digest",
                        "sha256:1111111111",
                        "sha256:1111111111",
                        "sha256:2222222222",
                        "sha256:2222222222",
                        "sha256:3333333333",
                        "sha256:3333333333",
                        "sha256:5555555555",
                        "sha256:5555555555",
                    ],
                    identity=[
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                        "some-registry1.com/target/repo:latest-test-tag",
                        "some-registry2.com/target/repo:latest-test-tag",
                    ],
                ),
            ]
        )



import mock
import pytest
import requests
import requests_mock

from pubtools._quay import quay_api_client


@mock.patch("pubtools._quay.quay_api_client.QuaySession")
def test_init(mock_session):
    client = quay_api_client.QuayApiClient("some-token", "stage.quay.io")

    assert client.token == "some-token"
    mock_session.assert_called_once_with(hostname="stage.quay.io", api="quay")
    mock_session.return_value.set_auth_token.assert_called_once_with("some-token")


def test_delete_client():
    client = quay_api_client.QuayApiClient("some-token", "stage.quay.io")

    with requests_mock.Mocker() as m:
        m.delete(
            "https://stage.quay.io/api/v1/repository/some-repo/tag/10",
            [
                {"text": "Unauthorized", "status_code": 401},
                {"text": "Success", "status_code": 200},
                {"text": "Not found", "status_code": 404},
            ],
        )
        with pytest.raises(requests.HTTPError, match="401 Client Error.*"):
            resp = client.delete_tag("some-repo", "10")

        resp = client.delete_tag("some-repo", "10")
        assert resp.status_code == 200
        resp = client.delete_tag("some-repo", "10")
        assert resp.text == "Not found"
        assert resp.status_code == 404

        assert m.call_count == 3


def test_delete_repository():
    client = quay_api_client.QuayApiClient("some-token", "stage.quay.io")

    with requests_mock.Mocker() as m:
        m.delete(
            "https://stage.quay.io/api/v1/repository/some-namespace/some-repo",
            [
                {"text": "Server error", "status_code": 500},
                {"text": "Success", "status_code": 200},
            ],
        )
        with pytest.raises(requests.HTTPError, match="500 Server Error.*"):
            client.delete_repository("some-namespace/some-repo")

        response = client.delete_repository("some-namespace/some-repo")
        assert response.status_code == 200

        assert m.call_count == 2



class ManifestTypeError(Exception):
    """Occurs when an incorrect manifest type is encountered."""


class RegistryAuthError(Exception):
    """Occurs when registry authentication encounters an issue."""


class BadPushItem(Exception):
    """Occurs when a bad push item is being processed."""


class InvalidTargetSettings(Exception):
    """Occurs when required target setting is missing or has an incorrect value."""


class InvalidRepository(Exception):
    """Occurs when a repository hasn't passed its validation checks."""


class SigningError(Exception):
    """Occurs when there was an issue in the container signing process done by RADAS."""


class ManifestNotFoundError(Exception):
    """Occurs when manifest is not found on the server."""



from __future__ import annotations

import logging
import requests
import tempfile
from collections.abc import Iterable

from .quay_client import QuayClient
from .quay_api_client import QuayApiClient
from .security_manifest_pusher import SecurityManifestPusher
from .types import Manifest, ManifestList

from typing import cast, List, Dict, Optional, Tuple, Union

LOG = logging.getLogger("pubtools.quay")


class ImageUntagger:
    """Class containing logic for untagging images and deciding if they should be untagged."""

    def __init__(
        self,
        references: List[str],
        quay_api_token: str,
        remove_last: bool = False,
        quay_user: Optional[str] = None,
        quay_password: Optional[str] = None,
        host: Optional[str] = None,
    ) -> None:
        """
        Initialize.

        Args:
            references ([str]):
                List of image references to untag.
            quay_api_token (str):
                OAuth token for authentication of Quay REST API.
            remove_last (bool):
                Whether to remove a tag when it's the last reference of an image (in that repo).
            quay_user (str):
                Quay username for Docker HTTP API.
            quay_password (str):
                Quay password for Docker HTTP API.
            host (str):
                Custom Quay hostname (if required).
        """
        for reference in references:
            if "@" in reference:
                raise ValueError(
                    "Reference '{0}' must be specified via tag, not digest".format(reference)
                )

        if isinstance(host, str) and host[-1] == "/":
            host = host[:-1]
        self.host = host or "quay.io"

        self.references = references
        self.remove_last = remove_last

        if quay_user and quay_password:
            self._quay_client: Optional[QuayClient] = QuayClient(
                quay_user, quay_password, self.host
            )
        else:
            self._quay_client = None
        self._quay_api_client = QuayApiClient(quay_api_token, self.host)

    def set_quay_client(self, quay_client: QuayClient) -> None:
        """
        Set client instance to be used for the HTTP API operations.

        Args:
            quay_client (QuayClient):
                Instance of QuayClient.
        """
        self._quay_client = quay_client

    def get_repository_tags_mapping(self) -> Dict[str, List[str]]:
        """
        Get a mapping of which tags would be removed from given repos based on provided refs.

        Returns ({str: [str]}):
            Mapping of repository->tags.
        """
        repo_tag_mapping: Dict[str, List[str]] = {}
        for reference in self.references:
            tag = reference.split(":")[-1]
            repo_path = reference.split(":")[0]
            # Take only last two parts of the url, which are namespace + repo
            repository = "/".join(repo_path.split("/")[-2:])
            repo_tag_mapping.setdefault(repository, []).append(tag)

        return repo_tag_mapping

    def construct_tag_digest_mappings(
        self, repository: str
    ) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
        """
        Create a mappings of tags->digests as well as digests->tags.

        Tags->digests mapping answers the question "Which digests does a given tag reference?".
        Digests->tags mapping answers the question "Which tags reference a given digest?".
        Tag may reference one digest (if image manifest), or multiple (if manifest list).

        Args:
            repository (str):
                Quay repository.

        Returns (({str: [str]}), ({str: [str]})):
            Tuple of dictionaries mapping tags->digests and digests->tags.
        """
        LOG.info("Gathering tags and digests of repository '{0}'".format(repository))
        tag_digest_mapping: Dict[str, List[str]] = {}
        digest_tag_mapping: Dict[str, List[str]] = {}
        image_schema = "{0}/{1}:{2}"
        repo_tags = cast(
            Dict[str, List[str]],
            cast(QuayClient, self._quay_client).get_repository_tags(repository),
        )

        for tag in repo_tags["tags"]:
            image = image_schema.format(self.host, repository, tag)
            try:
                manifest = cast(
                    Union[ManifestList, Manifest],
                    cast(QuayClient, self._quay_client).get_manifest(image),
                )
            except requests.exceptions.HTTPError as e:
                # Just removed tags could still be in tags list while manifests are removed
                if e.response.status_code == 404:
                    continue
                else:
                    raise
            digest = cast(QuayClient, self._quay_client).get_manifest_digest(image)

            # Option 1: No manifest list, only manifest
            if manifest.get("mediaType", None) in (
                QuayClient.MANIFEST_V2S2_TYPE,
                QuayClient.MANIFEST_OCI_V2S2_TYPE,
                QuayClient.MANIFEST_V2S1_TYPE,
                None,
            ):
                tag_digest_mapping[tag] = [digest]
                digest_tag_mapping.setdefault(digest, []).append(tag)

            # Option 2: We need to get digests of all architectures
            else:
                tag_digest_mapping[tag] = [digest]
                digest_tag_mapping.setdefault(digest, []).append(tag)

                for arch_manifest in cast(ManifestList, manifest)["manifests"]:
                    tag_digest_mapping[tag].append(arch_manifest["digest"])
                    digest_tag_mapping.setdefault(arch_manifest["digest"], []).append(tag)

        return (tag_digest_mapping, digest_tag_mapping)

    def get_lost_digests(
        self,
        tags: List[str],
        tag_digest_mapping: Dict[str, List[str]],
        digest_tag_mapping: Dict[str, List[str]],
    ) -> List[str]:
        """
        Calculate a list of digests that would be lost if the provided tags were removed.

        Args:
            tags ([str]):
                List of tags that would be removed from a repo.
            tag_digest_mapping ({str: [str]}):
                Mapping of which digests are referenced by a given tag.
            digest_tag_mapping ({str: [str]}):
                Mapping of which tags reference a given digest.

        Returns ([str]):
            Digests that would be lost if given tags were removed.
        """
        remove_digests = []
        lost_digests = []
        for tag in tags:
            for digest in tag_digest_mapping.get(tag, []):
                if digest not in remove_digests:
                    remove_digests.append(digest)

        for digest in remove_digests:
            # which tags would remain referencing a given digest?
            remaining_tags = set(digest_tag_mapping[digest]) - set(tags)
            if len(remaining_tags) == 0:
                lost_digests.append(digest)

        return lost_digests

    def get_repo_cosign_images(
        self,
        repo_images: List[str],
        repo_tags: List[str],
        image_types: Optional[Iterable[str]] = None,
    ) -> set[str]:
        """
        Get a list of images generated by cosign associated with the provided images.

        The provided images are expected to be specified by digest (as tag can point to
        multiple images). This method will return associated "secondary" images likely generated by
        cosign. They can be attestations, sboms, or signatures. Their tag likely has format
        "sha256-<digest>.<sbom|att|sig>".

        Args:
            repo_images ([str]):
                Main images whose "cosign" images will be found.
            repo_tags ([str]):
                All tags in a repo.
            image_types ([str]):
                Which cosign image types to get (supported values are in COSIGN_TRIANGULATE_TYPES).
                If unset, all image types will be checked.
        Returns ([str]):
            List of cosign images associated with the provided images.
        """
        LOG.info(f"Getting cosign images of {len(repo_images)} images")
        if not repo_images:
            return set()

        check_types = (
            image_types if image_types else SecurityManifestPusher.COSIGN_TRIANGULATE_TYPES
        )
        types_found = [t in SecurityManifestPusher.COSIGN_TRIANGULATE_TYPES for t in check_types]
        if not all(types_found):
            raise ValueError(f"Unknown cosign image types in {image_types}")

        # check if all images are specified by digest
        nondigest_images = [i for i in repo_images if "@sha256:" not in i]
        if nondigest_images:
            raise ValueError(f"Images {nondigest_images} are not specified by digest")
        # check if all images belong to the same repo
        repos = set([i.split("@")[0] for i in repo_images])
        if len(repos) > 1:
            raise ValueError("Specified images belong to multiple repos")

        set_repo_tags = set(repo_tags)
        cosign_images = set()

        with tempfile.TemporaryDirectory(prefix="security_manifest_") as tmp_dir:
            for image in repo_images:
                for image_type in check_types:
                    cosign_image = SecurityManifestPusher.cosign_triangulate_image(
                        image, tmp_dir, image_type
                    )
                    if cosign_image.split(":")[-1] in set_repo_tags:
                        cosign_images.add(cosign_image)

        LOG.info(f"{len(cosign_images)} cosign images were found for the {len(repo_images)} images")
        return cosign_images

    def untag_images(self) -> List[str]:
        """
        Determine if the specified tags may be removed and remove them.

        Returns ([str]):
            List of image references lost by the untagging.
        """
        if not self._quay_client:
            raise RuntimeError("QuayClient instance must be set")

        repo_tag_mapping = self.get_repository_tags_mapping()
        schema = "{0}/{1}@{2}"
        lost_imgs = []
        lost_cosign_imgs = set()

        for repo in repo_tag_mapping:
            tag_digest_mapping, digest_tag_mapping = self.construct_tag_digest_mappings(repo)

            lost_digests = self.get_lost_digests(
                repo_tag_mapping[repo], tag_digest_mapping, digest_tag_mapping
            )
            if len(lost_digests) > 0:
                lost_repo_images = [
                    schema.format(self.host, repo, digest) for digest in lost_digests
                ]
                lost_imgs += lost_repo_images

                repo_tags: List[str] = cast(
                    Dict[str, List[str]], self._quay_client.get_repository_tags(repo)
                )["tags"]
                # We need to run this twice to get all secondary cosign images.
                # First, we find sbom, att, sig images of normal images.
                # Second, we find sig images of sbom and att images.
                repo_lost_cosign_imgs = self.get_repo_cosign_images(lost_repo_images, repo_tags)
                cosign_imgs_by_digest = sorted(
                    [
                        f"{i.split(':')[0]}@{self._quay_client.get_manifest_digest(i)}"
                        for i in repo_lost_cosign_imgs
                    ]
                )
                repo_lost_cosign_imgs |= self.get_repo_cosign_images(
                    cosign_imgs_by_digest, repo_tags, ["signature"]
                )
                lost_cosign_imgs |= repo_lost_cosign_imgs

        if not self.remove_last and lost_imgs:
            raise ValueError(
                "Following images would no longer be referencable by tag after"
                " the untagging operation: '{0}'. Please specify --remove-last"
                " if this is not a concern.".format(lost_imgs)
            )

        if lost_imgs and self.remove_last:
            LOG.warning("Following images won't be referencable by tag: {0}".format(lost_imgs))
        if not lost_imgs:
            LOG.info("No images will be lost by this untagging operation")
        if lost_cosign_imgs:
            LOG.warning(
                "Following cosign images won't be referecable by tag: "
                f"{sorted(list(lost_cosign_imgs))}"
            )

        for reference in sorted(list(set(self.references) | lost_cosign_imgs)):
            tag = reference.split(":")[-1]
            repo_path = reference.split(":")[0]
            # Take only last two parts of the url, which are namespace + repo
            repository = "/".join(repo_path.split("/")[-2:])
            LOG.info("Removing tag '{0}' from repository '{1}'".format(tag, repository))
            self._quay_api_client.delete_tag(repository, tag)

        return sorted(list(set(lost_imgs) | lost_cosign_imgs))



import argparse
import logging
import os
from typing import Any, Dict, List, cast, Optional

from pubtools.pluggy import task_context, pm

from .quay_client import QuayClient
from .untag_images import untag_images
from .utils.misc import (
    setup_arg_parser,
    add_args_env_variables,
    get_internal_container_repo_name,
)
from .item_processor import (
    item_processor_for_internal_data,
    VirtualPushItem,
)
from .signer_wrapper import SIGNER_BY_LABEL

LOG = logging.getLogger("pubtools.quay")

CLEAR_REPO_ARGS = {
    ("--repositories",): {
        "help": "External repositories to clear as CSV.",
        "required": True,
        "type": str,
    },
    ("--quay-host",): {
        "help": "Quay host name",
        "required": False,
        "type": str,
        "default": "quay.io",
    },
    ("--quay-org",): {
        "help": "Quay organization in which repositories reside.",
        "required": True,
        "type": str,
    },
    ("--quay-api-token",): {
        "help": "OAuth token for Quay REST API.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_API_TOKEN",
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": True,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--pyxis-server",): {
        "help": "Pyxis service hostname",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-crtfile",): {
        "help": "Path to .crt file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-keyfile",): {
        "help": "Path to .key file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--pyxis-request-threads",): {
        "help": "Maximum number of threads to use for parallel pyxis request",
        "required": False,
        "default": 7,
        "type": int,
    },
    ("--signers",): {
        "help": "Comma separated list of signers",
        "required": False,
        "type": str,
        "default": "",
    },
    ("--signer-configs",): {
        "help": "Comma separated list of paths to signer configs",
        "required": False,
        "type": str,
        "default": "",
    },
}


def clear_repositories(repositories: str, settings: Dict[str, Any]) -> None:
    """
    Clear Quay repository.

    Args:
        repository (str):
            External repositories to clear. Comma separated values.
        settings (dict):
            quay_org (str):
                Quay organization in which repositories reside.
            quay_api_token (str):
                OAuth token for authentication of Quay REST API.
            quay_user (str):
                Quay username for Docker HTTP API.
            quay_password (str):
                Quay password for Docker HTTP API.
            pyxis_server (str):
                Pyxis service hostname:
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.
            pyxis_request_threads:
                Maximum number of threads to use for parallel pyxis request.
    """
    parsed_repositories = repositories.split(",")

    LOG.info("Clearing repositories '{0}'".format(repositories))
    quay_client = QuayClient(settings["quay_user"], settings["quay_password"])
    item_processor = item_processor_for_internal_data(
        quay_client, "quay.io", [""], 5, settings["quay_org"]
    )
    item_processor.extractor.full_extract = True

    signer_settings = {k: v for k, v in settings.items() if k not in ["quay_org"]}
    signer_settings["quay_namespace"] = settings["quay_org"]
    signer_settings["dest_quay_api_token"] = os.environ.get("QUAY_API_TOKEN")
    signer_settings["quay_host"] = "quay.io"

    # Clear repository doesn't work with pushitem so we need to create a virtual push item
    # to use existing code to generate needed data for clearing the repository
    item = VirtualPushItem(
        metadata={"tags": {repo: [] for repo in parsed_repositories}},
        repos={repo: [] for repo in parsed_repositories},
    )
    existing_manifests = item_processor.generate_all_existing_manifests_metadata(item)
    signers = settings["signers"].split(",")
    signer_configs = settings["signer_configs"].split(",")
    outdated_manifests = []
    for repo, tag, mad in existing_manifests:
        if not mad:
            continue
        outdated_manifests.append((mad.digest, tag, repo))

    for n, signer in enumerate(signers):
        signercls = SIGNER_BY_LABEL[signer]
        if not signercls.pre_push:
            continue
        _signer = signercls(config_file=signer_configs[n], settings=settings)
        _signer.remove_signatures(outdated_manifests, _exclude=[])

    refrences_to_remove = []
    for repository in parsed_repositories:
        internal_repo = "{0}/{1}".format(
            settings["quay_org"], get_internal_container_repo_name(repository)
        )
        repo_data = cast(Dict[str, List[str]], quay_client.get_repository_tags(internal_repo))

        for tag in repo_data["tags"]:
            refrences_to_remove.append("{0}/{1}:{2}".format("quay.io", internal_repo, tag))

    untag_images(
        sorted(refrences_to_remove),
        settings["quay_api_token"],
        remove_last=True,
        quay_user=settings["quay_user"],
        quay_password=settings["quay_password"],
    )

    LOG.info("Repositories have been cleared")
    pm.hook.quay_repositories_cleared(repository_ids=sorted(parsed_repositories))


def setup_args() -> argparse.ArgumentParser:
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(CLEAR_REPO_ARGS)


def clear_repositories_main(sysargs: Optional[List[str]] = None) -> None:
    """Entrypoint for clearing repositories."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, CLEAR_REPO_ARGS)

    if not args.quay_api_token:
        raise ValueError("--quay-api-token must be specified")
    if not args.quay_password:
        raise ValueError("--quay-password must be specified")

    kwargs = args.__dict__
    repositories = kwargs.pop("repositories")

    with task_context():
        clear_repositories(repositories, kwargs)



import dataclasses
from typing import Dict, Any, List, Optional


@dataclasses.dataclass
class BuildIndexImageParam:
    """Parameter data for building index image and part of data required by iib_results."""

    bundles: List[str]
    index_image: str
    deprecation_list: Optional[List[str]]
    build_tags: List[str]
    target_settings: Dict[str, Any]
    tag: str
    signing_keys: List[str]
    destination_tags: List[str]
    origin: str
